import {Complex} from './type.js'

import {implementations, commonSpecs} from '@/core/Dispatcher'
import type {RawDependencies} from '@/core/Dispatcher'
import {commonSignature, RealType, $reflect} from '@/interfaces/type'
import type {CommonSignature, CommonReturn} from '@/interfaces/type'

// Narrowly typed signature selectors, for the operations we need to use
// with atypical signatures:
const add = 'add' as const
const divide = 'divide' as const

export function common<T>() {
   const baseSignature = commonSpecs<T>()
   const withComplex
      = <RD extends RawDependencies>(rd: RD) => baseSignature({
         ...rd, complex: {}
      })

   const realSignature = commonSpecs<RealType<T>>()

   return implementations<CommonSignature<Complex<T>>>()
      .dependent(withComplex({add: {}}), {
         add: dep => (w, z) =>
            dep.complex(dep.add(w.re, z.re), dep.add(w.im, z.im))
      })
      .dependent(withComplex({unaryMinus: {}}), {
         unaryMinus: dep => z =>
            dep.complex(dep.unaryMinus(z.re), dep.unaryMinus(z.im))
      })
      .dependent(withComplex({unaryMinus: {}, conj: {}}), {
         conj: dep => z => dep.complex(dep.conj(z.re), dep.unaryMinus(z.im))
      })
      .dependent(withComplex({subtract: {}}), {
         subtract: dep => (w, z) =>
            dep.complex(dep.subtract(w.re, z.re), dep.subtract(w.re, z.re))
      })
      .dependent(withComplex({add: {}, subtract: {}, multiply: {}, conj: {}}), {
         multiply: dep => (w, z) => {
            const mult = dep.multiply
            const realpart = dep.subtract(
               mult(         w.re,  z.re), mult(dep.conj(w.im), z.im))
            const imagpart = dep.add(
               mult(dep.conj(w.re), z.im), mult(         w.im,  z.re))
            return dep.complex(realpart, imagpart)
         }
      })
      .dependent(baseSignature({
         absquare: {},
         add: {sig: commonSignature<'add', CommonReturn<'absquare', T>>()}
      }), {
         absquare: dep => z => dep.add(dep.absquare(z.re), dep.absquare(z.im))
      })
      .dependent({
         conj: {},
         absquare: {},
         divideReal: {
            is: divide,
            sig: (a: Complex<T>, b: RealType<T>) => ({} as Complex<T>)
         }
      }, {
         reciprocal: dep => z => dep.divideReal(dep.conj(z), dep.absquare(z))
      })
      .dependent({multiply: {}, reciprocal: {}}, {
         divide: dep => (w,z) => dep.multiply(w, dep.reciprocal(z))
      })
      // The dependencies are tricky in the implementation of `sqrt` below,
      // because there are three types involved: Complex<T>, T, and
      // RealType<T>, all of which might be different.
      // We have to get it straight which operations we need on each type;
      // for example, we need `add` on three different combinations:
      .dependent({
         absquare: {}, re: {},       // Complex<T>-dependencies
         ...withComplex({zero: {}}), // T-dependencies
         // And RealType<T>-dependencies:
         ...realSignature({
             conservativeSqrt: {}, equal: {}, unaryMinus: {}, isnan: {}
         }),
         // And now mixed dependencies:
         divideReal: {
            is: divide,
            sig: (a: Complex<T>, b: RealType<T>) => ({} as Complex<T>)
         },
         addRR: {
            is: add,
            sig: (a: RealType<T>, b: RealType<T>) => ({} as RealType<T>)
         },
         addTR: {is: add, sig: (a: T, b: RealType<T>) => ({} as T)},
         addCR: {
            is: add,
            sig: (a: Complex<T>, b: RealType<T>) => ({} as Complex<T>)
         }
      }, {
         sqrt: dep => z => {
            const absq = dep.absquare(z)
            const myabs = dep.conservativeSqrt(absq)
            if (dep.isnan(myabs)) {
               throw new RangeError(
                  `sqrt(${z}): cannot take square root of norm square ${absq}`
               )
            }
            const r = dep.re(z)
            const negr = dep.unaryMinus(r)
            if (dep.equal(myabs, negr)) {
               // pure imaginary square root; z.im already zero
               const rootNegr = dep.conservativeSqrt(negr)
               if (dep.isnan(rootNegr)) {
                  throw new RangeError(
                     `sqrt(${z}): cannot take square root of `
                        + `negative real part ${negr}`
                  )
               }
               return dep.complex(
                  dep.zero(z.re), dep.addTR(z.im, rootNegr))
            }
            const num = dep.addCR(z, myabs)
            const denomsq = dep.addRR(dep.addRR(myabs, myabs), dep.addRR(r, r))
            const denom = dep.conservativeSqrt(denomsq)
            if (dep.isnan(denom)) {
               throw new RangeError(
                  `sqrt(z) for z = ${z}: cannot take square root of `
                     + `2|z| + 2re(z) = ${denomsq}`
               )
            }
            return dep.divideReal(num, denom)
         }
      })
      .ship()
}

// Additional implementations for non-uniform signatures
export function mixed<T>() {
   return implementations<{
      add: (z: Complex<T>, r: RealType<T>) => Complex<T>,
      divide: (z: Complex<T>, r: RealType<T>) => Complex<T>,
      complex: CommonSignature<T>['complex']
   }>()
      .dependent({
         addTR: {is: add, sig: (a: T, b: RealType<T>) => ({} as T)},
         complex: {}
      }, {
         add: dep => (z, r) => dep.complex(dep.addTR(z.re, r), z.im)
      })
      .dependent({
         divTR: {is: divide, sig: (a: T, b: RealType<T>) => ({} as T)},
         complex: {}
      }, {
         divide: dep => (z, r) =>
            dep.complex(dep.divTR(z.re, r), dep.divTR(z.im, r))
      })
      .ship()
}

$reflect!([common, mixed])
