import {$$typeToString} from 'ts-macros'

import {Configuration} from '@/core/Configuration'

/* First some type utilities: */

export type ToMapped<T> = {[K in keyof T]: T[K]}

export type UnionToIntersection<U> = 
   (U extends any ? (k: U) => void : never) extends ((k: infer I) => void)
      ? I : never 

export type ValueIntersectionByKeyUnion<T,  TKey extends keyof T> = {
   [P in TKey]: (k: T[P]) => void
} [TKey] extends ((k: infer I) => void) ? I : never


// The following needs to be a type that is extended by
// the type of any function. Hopefully I have gotten this right.
export type AnyFunc = (...args: never[]) => unknown

// An ensemble of signatures of operations, for example, the CommonSignatures
// collected up by all of the modules
export type GenSigs = Record<string, AnyFunc>

/*****
 * Every core math type has some associated types; they need
 * to be published in the following interface. The key is the
 * name of the type, and within the subinterface for that key,
 * the type of the 'type' property is the actual TypeScript type
 * we are associating the other properties to. To get an associated type,
 * types are looked up by matching this 'type' property.
 *
 * Note the interface is generic with one parameter. To actually find
 * the associated types of type T, instantiate the interface with type
 * T. This mechanism deals with the fact that TypeScript doesn't really
 * deal well with interfaces, some entries of which are generic and others
 * are not.
 ****/

export interface AssociatedTypes<T> {
   undefined: {
      type: undefined
      zero: undefined // The type of the zero of this type
      one: undefined  // The type of the multiplicative identity of this type
      nan: undefined  // The type of Not a Number of this type
      real: undefined // The type of the real part of this type
      closure: undefined // The type of the algebraic closure of this type
   }
}

export type TypeName = string // Really should be some recursive definition,
// any key of AssociatedTypes that's not generic, or any key that is generic
// but instantatied by a TypeName. Not sure how to do that, so just go with
// any string for now.

type AssociatedTypeNames = keyof AssociatedTypes<unknown>['undefined']
type ALookup<T, Name extends AssociatedTypeNames> =
   ValueIntersectionByKeyUnion<
      {[K in keyof AssociatedTypes<T>]:
         T extends AssociatedTypes<T>[K]['type']
            ? AssociatedTypes<T>[K][Name] : unknown
      },
      keyof AssociatedTypes<T>
   >

// For everything to compile, zero and one must be subtypes of T:
export type ZeroType<T> = ALookup<T, 'zero'> & T
export type OneType<T> = ALookup<T, 'one'> & T
// But I believe 'nan' really might not be, like I think we will have to use
// number NaN for the nan of 'bigint', as it has nothing at all like NaN,
// so don't force it:
export type NaNType<T> = ALookup<T, 'nan'>
export type RealType<T> = ALookup<T, 'real'>
export type ClosureType<T> = ALookup<T, 'closure'>

/*****
 * The typical signature for every operation needs to be published in the
 * following interface. Each key is the name of an operation.
 * The type of each key should be the function type that the operation would
 * "normally" have, understanding that there may be exceptions, which will\
 * be dealt with by another mechanism.
 *
 * Note that this interface is generic in two parameters, the second of which
 * defaults to the first. These are slots for type parameters to the typical
 * signatures, most of which have only one type parameter.
 ****/

export interface CommonInterface<T, Aux = T> {
    zero: (a: T) => ZeroType<T>
    one: (a: T) => OneType<T>
    nan: (a: T | NaNType<T>) => NaNType<T>
    re: (a: T) => RealType<T>

    config: () => Configuration
    convert: (from: T) => Aux
    typeOf: (x: unknown) => TypeName
}

export type CommonSignature<T, Aux = T> = ToMapped<CommonInterface<T, Aux>>
export type SignatureKey<T, Aux = T> = keyof CommonSignature<T, Aux>

export function commonSignature<
   K, T, Aux = T, CS extends GenSigs = CommonSignature<T, Aux>
>(): K extends keyof CS ? CS[K] : () => void {
   return (() => undefined) as K extends keyof CS ? CS[K] : () => void
}

export type CommonReturn<
   K, T, Aux = T, CS extends GenSigs = CommonSignature<T, Aux>
> = K extends keyof CS ? ReturnType<CS[K]> : void

export type Dependency<T, Aux = T> = {
   [K in SignatureKey<T, Aux>]: Pick<CommonSignature<T, Aux>, K>
}

export type Dependencies<T, Names extends SignatureKey<T>, Aux = T> =
    Pick<CommonSignature<T, Aux>, Names>

// Macro for type reflection:
export function $reflect<ImplTuple>(tup: ImplTuple) {
   +[[tup], <T>(elt: T) => {
      (elt as {reflectedType: string}).reflectedType
        = $$typeToString!<T>(true, false, true);
   }]
}
